package main

import (
	"context"
	"encoding/json"
	"errors"
	"flag"
	"fmt"
	"io/fs"
	"os"
	"sort"
	"strings"

	"github.com/adrg/xdg"
	"github.com/peterbourgon/ff/v3"
	"github.com/peterbourgon/ff/v3/ffcli"
)

var Version = "develop"

func main() {
	defaultStore, err := xdg.ConfigFile("kv/store.json")
	if err != nil {
		defaultStore = "store.json"
	}

	fs := flag.NewFlagSet("kv", flag.ContinueOnError)
	storeFlag := fs.String("store", defaultStore, "Location of the store on disk")
	fs.StringVar(storeFlag, "s", *storeFlag, "--store")

	a := &app{
		storeLocation: storeFlag,
	}

	var app *ffcli.Command
	app = &ffcli.Command{
		Name:       "kv",
		ShortUsage: fmt.Sprintf("Key/Value store - version %s", Version),
		ShortHelp:  "kv [get|set|del|list] ...",
		FlagSet:    fs,
		Subcommands: []*ffcli.Command{
			a.get(),
			a.set(),
			a.del(),
			a.list(),
		},
		Options: []ff.Option{
			ff.WithEnvVarPrefix("KV"),
		},
		Exec: func(_ context.Context, _ []string) error {
			return errors.New(app.UsageFunc(app))
		},
	}

	if err := app.ParseAndRun(context.Background(), os.Args[1:]); err != nil {
		if errors.Is(err, flag.ErrHelp) {
			return
		}
		fmt.Println(err)
	}
}

type app struct {
	storeLocation *string
}

func (a *app) get() *ffcli.Command {
	fs := flag.NewFlagSet("get", flag.ContinueOnError)
	return &ffcli.Command{
		Name:    "get",
		FlagSet: fs,
		Exec: func(ctx context.Context, args []string) error {
			if len(args) < 1 {
				return errors.New("get requires at least one argument")
			}

			data, err := a.load()
			if err != nil {
				return err
			}

			key := strings.ToLower(strings.Join(args, " "))
			val, ok := data[key]
			if !ok {
				return fmt.Errorf("no value found for %q", key)
			}
			fmt.Print(val)
			return nil
		},
	}
}

func (a *app) set() *ffcli.Command {
	fs := flag.NewFlagSet("set", flag.ContinueOnError)
	return &ffcli.Command{
		Name:    "set",
		FlagSet: fs,
		Exec: func(ctx context.Context, args []string) error {
			if len(args) < 2 {
				return errors.New("set requires at least two arguments")
			}

			data, err := a.load()
			if err != nil {
				return err
			}

			key := args[0]
			data[key] = strings.Join(args[1:], " ")
			if err := a.save(data); err != nil {
				return err
			}
			fmt.Printf("set %q\n", key)
			return nil
		},
	}
}

func (a *app) del() *ffcli.Command {
	fs := flag.NewFlagSet("del", flag.ContinueOnError)
	return &ffcli.Command{
		Name:    "del",
		FlagSet: fs,
		Exec: func(ctx context.Context, args []string) error {
			if len(args) < 1 {
				return errors.New("del requires at least 1 argument")
			}

			data, err := a.load()
			if err != nil {
				return err
			}

			key := strings.ToLower(strings.Join(args, " "))
			if _, ok := data[key]; !ok {
				return fmt.Errorf("no value found for %q", key)
			}
			delete(data, key)
			if err := a.save(data); err != nil {
				return err
			}

			fmt.Printf("deleted %q\n", key)

			return nil
		},
	}
}

func (a *app) list() *ffcli.Command {
	fs := flag.NewFlagSet("list", flag.ContinueOnError)
	return &ffcli.Command{
		Name:    "list",
		FlagSet: fs,
		Exec: func(ctx context.Context, args []string) error {
			data, err := a.load()
			if err != nil {
				return err
			}

			var keys []string
			prefix := strings.ToLower(strings.Join(args, " "))
			for key := range data {
				if strings.HasPrefix(key, prefix) {
					keys = append(keys, key)
				}
			}
			sort.Strings(keys)

			for _, key := range keys {
				fmt.Println(key)
			}

			return nil
		},
	}
}

func (a *app) load() (map[string]string, error) {
	var m map[string]string
	fi, err := os.Open(*a.storeLocation)
	if err != nil {
		if errors.Is(err, fs.ErrNotExist) {
			fi, err = os.Create(*a.storeLocation)
			if err != nil {
				return m, err
			}
			if _, err := fi.WriteString("{}"); err != nil {
				return m, err
			}
			if err := fi.Close(); err != nil {
				return m, err
			}
			return a.load()
		}
		return m, err
	}
	defer fi.Close()
	return m, json.NewDecoder(fi).Decode(&m)
}

func (a *app) save(m map[string]string) error {
	fi, err := os.Create(*a.storeLocation)
	if err != nil {
		return err
	}
	defer fi.Close()

	enc := json.NewEncoder(fi)
	enc.SetIndent("", "\t")
	return enc.Encode(&m)
}
